__all__ = [
    'Module',
    'ModuleList',
    'ModuleDict',
    'Sequential',
    'get_rank',
    'get_local_rank',
    'get_world_size',
]

import os
from typing import TYPE_CHECKING

import torch.distributed as dist

if TYPE_CHECKING:
    import torch.nn

    class Module(torch.nn.Module):
        ...

    class ModuleDict(torch.nn.ModuleDict):
        ...

    class ModuleList(torch.nn.ModuleList):
        ...

    class Sequential(torch.nn.Sequential):
        ...

else:
    try:
        from mmcv.runner import BaseModule as Module
        from mmcv.runner import ModuleDict, ModuleList, Sequential
    except Exception:
        from torch.nn import Module, ModuleDict, ModuleList, Sequential


def get_rank(*args, **kwargs) -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_rank(*args, **kwargs)
    return 0


def get_local_rank() -> int:
    return int(os.getenv('LOCAL_RANK', -1))


def get_world_size(*args, **kwargs) -> int:
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size(*args, **kwargs)
    return 1



__all__ = [
    'iter_initialized',
    'init_iter',
    'get_iter',
    'globals_',
    'inc_iter',
]

from typing import Optional

from ._extensions import get_logger
from .configs import Config

globals_ = Config()


def iter_initialized() -> bool:
    return '_iter' in globals_


def init_iter(iter_: Optional[int] = 0) -> None:
    if iter_ is None:
        globals_.pop('_iter', None)
        return
    if '_iter' in globals_:
        get_logger().warning(
            f"iter={globals_._iter} has been reset to {iter_}.",
        )
    globals_._iter = iter_


def get_iter() -> int:
    return globals_._iter


def inc_iter() -> None:
    globals_._iter += 1



__all__ = [
    'LR_SCHEDULERS',
    'OPTIMIZERS',
    'build_param_group',
    'build_param_groups',
    'build_optimizer',
]

import inspect
import re
from functools import partial
from typing import Any, Dict, List, Optional, Sequence

import torch.nn as nn
import torch.optim as optim
import torch.optim.lr_scheduler as lr_scheduler

from ..base import Config, Registry, default_build


def build_param_group(
    model: nn.Module,
    params: Dict[str, Any],
) -> Dict[str, Any]:
    params = dict(params)
    regex = re.compile(params['params'])
    params['params'] = []
    for n, p in model.named_parameters():
        if regex.match(n):
            params['params'].append(p)
    return params


def build_param_groups(
    model: nn.Module,
    params: Optional[Sequence[Dict[str, Any]]] = None,
) -> List[Dict[str, Any]]:
    if params is None:
        return [dict(params=model.parameters())]
    builder = partial(build_param_group, model)
    return list(map(builder, params))


def build_optimizer(
    registry: Registry[optim.Optimizer],
    config: Config,
) -> optim.Optimizer:
    model = config.pop('model')
    config.params = build_param_groups(model, config.get('params', None))
    return default_build(registry, config)


OPTIMIZERS = Registry(
    'optimizers',
    base=optim.Optimizer,
    build_func=build_optimizer,
)
for _, class_ in inspect.getmembers(optim, inspect.isclass):
    assert issubclass(class_, optim.Optimizer)
    OPTIMIZERS.register(class_)

LR_SCHEDULERS: Registry[lr_scheduler._LRScheduler] = Registry(
    'lr_schedulers',
    base=lr_scheduler._LRScheduler,
)
for _, class_ in inspect.getmembers(optim.lr_scheduler, inspect.isclass):
    if issubclass(class_, lr_scheduler._LRScheduler):
        LR_SCHEDULERS.register(class_)



import string

from .base import ADAPTS, BaseAdapt


@ADAPTS.register_module()
class Custom(BaseAdapt):
    """Custom adaptation described using patterns.

    Examples:
        >>> adapt = Custom(pattern='a + b * c')
        >>> adapt(1, 2, 3)
        7
        >>> adapt = Custom(pattern='a + var1 * var2')
        >>> adapt(1, var1=2, var2=3)
        7
        >>> adapt = Custom(pattern='a + b * c')
        >>> adapt(1, 2, 3, b=4)
        Traceback (most recent call last):
            ...
        RuntimeError: {'b'}
    """

    def __init__(self, *args, pattern: str, **kwargs):
        super().__init__(*args, **kwargs)
        self._pattern = pattern

    def forward(self, *args, **kwargs):
        locals_ = dict(zip(string.ascii_letters, args))
        if len(locals_.keys() & kwargs.keys()) != 0:
            raise RuntimeError(locals_.keys() & kwargs.keys())
        locals_.update(kwargs)
        return eval(self._pattern, None, locals_)



__all__ = [
    'get_modules',
    'no_grad',
    'eval_',
    'freeze',
    'FrozenMixin',
]

from typing import Any, List, Optional, Sequence, Tuple, TypeVar

import torch.nn as nn
from torch.nn.modules.batchnorm import _BatchNorm

from ..base import getattr_recur

DEFAULT_STOCHASTIC_MODULE_TYPES: Tuple[Any, ...] = (_BatchNorm, nn.Dropout)

try:
    from timm.models.layers import DropPath

    DEFAULT_STOCHASTIC_MODULE_TYPES += DropPath,
except Exception:
    pass


def get_modules(
    model: nn.Module,
    names: Sequence[str] = tuple(),
    types: Sequence[type] = tuple(),
) -> List[nn.Module]:
    if not names and not types:
        return [model]
    if not types:
        return [getattr_recur(model, name) for name in names]
    types = tuple(types)
    if not names:
        return [
            module for module in model.modules() if isinstance(module, types)
        ]
    modules = [getattr_recur(model, name) for name in names]
    modules = [module for module in modules if isinstance(module, types)]
    return modules


def no_grad(model: nn.Module, **kwargs) -> None:
    for module in get_modules(model, **kwargs):
        module.requires_grad_(False)


def eval_(model: nn.Module, **kwargs) -> None:
    for module in get_modules(model, **kwargs):
        module.eval()


def freeze(model: nn.Module, **kwargs) -> None:
    no_grad(model, **kwargs)
    eval_(model, **kwargs)


T = TypeVar('T', bound='FrozenMixin')


class FrozenMixin(nn.Module):

    def __init__(
        self,
        no_grad_config: Optional[dict] = None,
        eval_config: Optional[dict] = None,
    ) -> None:
        self._no_grad = no_grad_config
        self._eval = eval_config
        if no_grad_config:
            no_grad(self, **no_grad_config)
        if eval_config:
            eval_(self, **eval_config)

    def requires_grad_(self: T, requires_grad: bool = True) -> T:
        result = super().requires_grad_(requires_grad)
        if self._no_grad:
            no_grad(self, **self._no_grad)
        return result

    def train(self: T, mode: bool = True) -> T:
        result = super().train(mode)
        if self._eval:
            eval_(self, **self._eval)
        return result



__all__ = [
    'DebugMode',
    'BaseDebug',
]

import os

import torch.cuda

from ._extensions import get_logger


class DebugMode:

    def __init__(self) -> None:
        self._logger = get_logger()

    def __set_name__(self, owner, name: str) -> None:
        self._name = name
        self._forced = name in os.environ
        self._logger.debug(
            f"Debug mode {name} is {self.__get__(None)}"
            + (" (forced)" if self._forced else "")
        )

    def __get__(self, obj, objtype=None) -> bool:
        return bool(os.getenv(self._name))

    def __set__(self, obj, value: bool) -> None:
        if self._forced:
            self._logger.debug(
                f"Trying to set debug mode {self._name}, "
                f"which is forced to {self.__get__(None)}."
            )
            return
        if value:
            os.environ[self._name] = '1'
        else:
            os.environ.pop(self._name)
        self._logger.debug(
            f"Debug mode {self._name} is set to {self.__get__(None)}.",
        )


class BaseDebug:
    CPU = DebugMode()

    def init_cuda(self, **kwargs) -> None:
        assert not self.CPU

    def init_cpu(self, **kwargs) -> None:
        self.CPU = True

    def init_custom(self, **kwargs) -> None:
        pass

    def init(self, **kwargs) -> None:
        if torch.cuda.is_available():
            self.init_cuda(**kwargs)
        else:
            self.init_cpu(**kwargs)
        self.init_custom(**kwargs)



__all__ = [
    'DISTILLERS',
    'BaseDistiller',
    'DistillableProto',
    'build_metaclass',
]

import warnings
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Protocol,
    Set,
    Tuple,
    Type,
    TypeVar,
    cast,
)

import torch
import torch.nn as nn

from ..base import (
    Message,
    Module,
    ModuleList,
    Registry,
    Workflow,
    WorkflowConfig,
    transfer_weights,
)
from ..base.workflows import WorkflowSpec
from ..hooks import BaseHook

T = TypeVar('T', bound='BaseDistiller')


class BaseDistiller(Module):
    DEBUG_PREFIX = '_debug_'

    @staticmethod
    def workflow_to_module(workflow: Workflow) -> nn.Module:
        modules = ModuleList()
        for job in workflow:
            step_manager = job.step_manager
            if hasattr(step_manager, 'step'):
                module = step_manager.step  # type: ignore[attr-defined]
            elif hasattr(step_manager, 'steps'):
                module = ModuleList(
                    step_manager.steps,  # type: ignore[attr-defined]
                )
            else:
                raise TypeError(
                    f'{step_manager} has no attribute `step` or `steps`.'
                )
            modules.append(module)
        return modules

    def __init__(
        self,
        models: List[nn.Module],
        *,
        hooks: Dict[int, WorkflowConfig],
        adapts: WorkflowConfig,
        losses: WorkflowConfig,
        weight_transfer: Optional[Dict[str, str]] = None,
    ):
        Module.__init__(self)
        self._models = models

        self._hookflows: Dict[int, Workflow] = {  # yapf: disable
            i: Workflow.build('hooks', hooks_config)
            for i, hooks_config in hooks.items()
        }
        self._adaptflow = Workflow.build('adapts', adapts)
        self._lossflow = Workflow.build('losses', losses)

        if weight_transfer is not None:
            transfer_weights(self, weight_transfer)

        for i, workflow in self._hookflows.items():
            for job in workflow:
                for step in job:
                    cast(BaseHook, step).bind(self._models[i])

        self.add_module(
            '_adapts',
            self.workflow_to_module(self._adaptflow),
        )
        self.add_module(
            '_losses',
            self.workflow_to_module(self._lossflow),
        )

    @property
    def models(self) -> List[nn.Module]:
        return self._models

    def _apply(self: T, fn: Callable[..., None]) -> T:
        for model in self._models:
            if getattr(model, 'sync_apply', True):
                model._apply(fn)
        return super()._apply(fn)

    def hookflows(self) -> Iterator[Workflow]:
        return iter(self._hookflows.values())

    def hooks(self) -> Generator[BaseHook, None, None]:
        for workflow in self.hookflows():
            for job in workflow:
                yield from job

    def track_tensors(self) -> None:
        for hook in filter(
            lambda hook: hook.tracking_mode,
            self.hooks(),
        ):
            hook.track_tensor()

    def tensors(self) -> Message:
        tensors: Message = dict()
        for job in self.hookflows():
            job(tensors)
        return tensors

    def reset(self) -> None:
        for hook in self.hooks():
            hook.reset()

    def spec(self) -> WorkflowSpec:
        inputs: Set[str] = set()
        outputs: Set[str] = set()
        for i, hookflow in self._hookflows.items():
            hookflow_spec = hookflow.spec()
            assert len(hookflow_spec.inputs) == 0
            assert outputs.isdisjoint(hookflow_spec.outputs)
            outputs |= hookflow_spec.outputs
        adaptflow_spec = self._adaptflow.spec()
        inputs |= adaptflow_spec.inputs - outputs
        outputs |= adaptflow_spec.outputs
        lossflow_spec = self._lossflow.spec()
        inputs |= lossflow_spec.inputs - outputs
        outputs |= lossflow_spec.outputs
        return WorkflowSpec(inputs, outputs)

    def distill(
        self,
        custom_tensors: Optional[Dict[str, torch.Tensor]] = None,
        debug: bool = False,
    ) -> Dict[str, torch.Tensor]:
        if debug:
            expected_inputs = self.spec().inputs
            custom_inputs = (
                set()
                if custom_tensors is None else set(custom_tensors.keys())
            )
            if len(expected_inputs ^ custom_inputs):
                warnings.warn(
                    f"Missing inputs {expected_inputs - custom_inputs}\n"
                    f"Unexpected inputs {custom_inputs - expected_inputs}\n"
                )

        tensors = self.tensors()
        if custom_tensors is not None:
            tensors.update(custom_tensors)
        self._adaptflow(tensors)
        losses = self._lossflow(tensors.copy())

        if debug:
            losses.update({
                self.DEBUG_PREFIX + k: v
                for k, v in tensors.items()
            })
        return losses


DISTILLERS: Registry[BaseDistiller] = Registry(
    'distillers',
    base=BaseDistiller,
)


class DistillableProto(Protocol):
    _distiller: BaseDistiller

    @property
    def distiller(self) -> BaseDistiller:
        ...

    @property
    def sync_apply(self) -> bool:
        ...


def build_metaclass(
    distiller_cls: Type[BaseDistiller],
    supermetaclass: Type[type] = type,
) -> type:

    class MetaClass(supermetaclass):  # type: ignore[valid-type, misc]
        NAMESPACE = dict(
            distiller=property(
                lambda x: cast(DistillableProto, x)._distiller,
            ),
            sync_apply=property(lambda _: False),
        )

        def __new__(
            meta_cls,
            cls: str,
            bases: Tuple[type, ...],
            namespace: Dict[str, Any],
            **kwargs: Any,
        ) -> 'MetaClass':
            assert len(namespace.keys() & meta_cls.NAMESPACE.keys()) == 0
            namespace.update(meta_cls.NAMESPACE)
            return super().__new__(
                meta_cls,
                cls,
                bases,
                namespace,
                **kwargs,
            )

        def __call__(
            cls,
            *args,
            **kwargs,
        ) -> DistillableProto:
            if 'distiller' not in kwargs:
                raise RuntimeError('`distiller` is required')
            distiller: Dict[str, Any] = kwargs.pop('distiller')
            obj = super().__call__(*args, **kwargs)
            obj = cast(DistillableProto, obj)
            obj._distiller = distiller_cls(  # type: ignore[call-arg]
                student=obj,
                **distiller,
            )
            return obj

    return MetaClass
